<?php

/**
 * Fornece dados da pilha de execuo.
 * 
 * Supplies execution stack data.
 * 
 * Manipula e fornece acesso aos dados de uma determinada pilha de execuo. Por
 * meio de mtodos  possvel caminhar entre os nveis e capturar informaes
 * respectivas a cada um deles.
 * 
 * @author Renan de Lima Barbosa <renandelima@gmail.com>
 */
class BackTrace
{
	
	/**
	 * Indica que o escopo  de arquivo.
	 * 
	 * Indicates the scope is an file.
	 * 
	 *  um dos possveis valores retornado pelo mtodo
	 * {@link BackTrace::getScope()}. Indica que o ponteiro de execuo faz
	 * referncia  uma instruo cujo escopo  um arquivo.
	 * 
	 * @var string
	 */
	const SCOPE_FILE = 'file';
	
	/**
	 * Indica que o escopo  de funo.
	 * 
	 * Indicates the scope is an function.
	 * 
	 *  um dos possveis valores retornado pelo mtodo
	 * {@link BackTrace::getScope()}. Indica que o ponteiro de execuo faz
	 * referncia  uma instruo localizada em uma funo. Para verificar se o
	 * escopo  uma funo de classe utilize as constantes
	 * {@link BackTrace::SCOPE_METHOD} e {@link BackTrace::SCOPE_STATIC}.
	 * 
	 * @var string
	 */
	const SCOPE_FUNCTION = 'function';
	
	/**
	 * Indica que o escopo  de mtodo.
	 * 
	 * Indicates the scope is an method.
	 * 
	 *  um dos possveis valores retornado pelo mtodo
	 * {@link BackTrace::getScope()}. Indica que o ponteiro de execuo faz
	 * referncia  uma instruo localizada em um mtodo. Para verificar se o
	 * escopo  um mtodo esttico utilize a constante
	 * {@link BackTrace::SCOPE_STATIC}.
	 * 
	 * @var string
	 */
	const SCOPE_METHOD = '->';
	
	/**
	 * Indica que o escopo  de mtodo esttico.
	 * 
	 * Indicates the scope is an static method.
	 * 
	 *  um dos possveis valores retornado pelo mtodo
	 * {@link BackTrace::getScope()}. Indica que o ponteiro de execuo faz
	 * referncia  uma instruo localizada em um mtodo esttico. Para
	 * verificar se o escopo  um mtodo no esttico utilize a constante
	 * {@link BackTrace::SCOPE_METHOD}.
	 * 
	 * @var string
	 */
	const SCOPE_STATIC = '::';
	
	/**
	 * Contm os dados do nvel corrente.
	 * 
	 * Contains data of current level.
	 * 
	 * Ponteiro para um item da pilha {@link BackTrace::$aStack}. Os mtodos
	 * {@link BackTrace::levelDown()} e {@link BackTrace::levelUp()} so
	 * responsveis por manipular este atributo.
	 * 
	 * @var &string[]
	 */
	private $aCurrent = array();
	
	/**
	 * Contm os dados de todos os nveis.
	 * 
	 * Contains data of all levels.
	 * 
	 * Possui uma pilha com os dados de todos os nveis do rastreamento. O
	 * contedo deste atributo  o mesmo retornado pela funo
	 * {@link debug_backtrace()}.
	 * 
	 * @var string[][]
	 */
	private $aStack = array();
	
	/**
	 * Indicador do nvel atual.
	 * 
	 * Pointer of current level.
	 * 
	 * Contm o ndice do escopo atual. O valor desta varivel  um ndice de
	 * {@link BackTrace::$aStack}.
	 * 
	 * @var integer
	 */
	private $iLevel = 0;
	
	/**
	 * Captura dados da pilha de execuo.
	 * 
	 * Captures execution stack data.
	 * 
	 * Os dados da pilha de execuo so capturados e o ponteiro  movido para o
	 * nvel imediatamente inferior ao ponto de chamada deste construtor caso a
	 * pilha no seja informado pelo parmetro, caso contrrio a pilha informada
	 *  utilizada.
	 * 
	 * @param string[][] $aStack
	 * @return void
	 */
	public function __construct( array $aStack = array() )
	{
		$this->aStack = count( $aStack ) === 0 ? debug_backtrace() : $aStack;
		$this->levelTop();
	}
	
	/**
	 * Gera HTML com detalhes do rastreamento a partir do nvel atual.
	 * 
	 * Creates HTML with detail of the trace.
	 * 
	 * Captura todos os dados de um rastreamento a partir do nvel atual da
	 * pilha.
	 * 
	 * @return string
	 * @see BackTraceExplain::perform()
	 */
	public function explain()
	{
		return BackTraceExplain::perform( $this );
	}
	
	/**
	 * Obtm todos os dados de todos os nveis.
	 * 
	 * Gets all data of all levels.
	 * 
	 * Retorna os dados contidos na pilha construda pelo construtor dessa
	 * classe. O formato  idntico ao retornado pela funo
	 * {@link debug_backtrace()}.
	 * 
	 * @return string[]
	 */
	public function getAll()
	{
		return $this->aStack;
	}
	
	/**
	 * Obtm os argumentos de entrada do nvel atual.
	 * 
	 * Captures input arguments of current level.
	 * 
	 * Se o escopo do nvel corrente for uma funo ou mtodo retorna uma lista
	 * com os parmetros de chamada. Caso contrrio retorna uma lista vazia.
	 * 
	 * @return string[]
	 */
	public function getArguments()
	{
		if ( array_key_exists( 'args', $this->aCurrent ) === false )
		{
			return array();
		}
		return (array) $this->aCurrent['args'];
	}
	
	/**
	 * Obtm o nome da classe do nvel atual.
	 * 
	 * Captures class name of current level.
	 * 
	 * Se o escopo do nvel atual for um mtodo ou mtodo esttico o nome da
	 * classe  retornado. Caso contrrio retornar um texto vazio. Para
	 * verificar se o escopo atual  um mtodo, mtodo esttico, funo ou
	 * arquivo utilize {@link BackTrace::getScope()}.
	 * 
	 * @return string
	 */
	public function getClass()
	{
		if ( array_key_exists( 'class', $this->aCurrent ) === false )
		{
			return '';
		}
		return (string) $this->aCurrent['class'];
	}
	
	/**
	 * Obtm todos os dados do nvel corrente.
	 * 
	 * Captures all data of current level.
	 *
	 * Retorna os dados do nvel corrente da pilha de execuo. O formato 
	 * igual a um item da pilha retornada pela funo {@link debug_backtrace()}.
	 *
	 * @return string[]
	 */
	public function getCurrent()
	{
		return $this->aCurrent;
	}
	
	/**
	 * Obtm o nome do arquivo do nvel corrente.
	 * 
	 * Captures file of current level.
	 * 
	 * Retorna o caminho completo contendo o nome do arquivo do nvel atual.
	 * 
	 * @param boolean $bOnlyName captura somente o nome do arquivo
	 * @return string
	 */
	public function getFile( $bOnlyName = false )
	{
		if ( array_key_exists( 'file', $this->aCurrent ) === false )
		{
			return '';
		}
		if ( $bOnlyName === true )
		{
			return basename( (string) $this->aCurrent['file'] );
		}
		return (string) $this->aCurrent['file'];
	}
	
	/**
	 * Obtm o nome da funo do nvel corrente.
	 * 
	 * Captures function name of current level.
	 * 
	 * Se o escopo do nvel atual for uma funo, mtodo ou mtodo esttico o
	 * nome desse  retornado. Caso contrrio retornar um texto vazio. Para
	 * verificar se o escopo atual  uma funo, mtodo, mtodo esttico ou
	 * arquivo utilize {@link BackTrace::getScope()}.
	 * 
	 * @return string
	 */
	public function getFunction()
	{
		if ( array_key_exists( 'function', $this->aCurrent ) === false )
		{
			return '';
		}
		return (string) $this->aCurrent['function'];
	}
	
	/**
	 * Obtm a posio do nvel atual do nvel corrente.
	 * 
	 * Captures position of current level.
	 * 
	 * @return integer
	 */
	public function getLevel()
	{
		return $this->iLevel;
	}
	
	/**
	 * Obtm nmero da linha do arquivo do nvel corrente.
	 * 
	 * Captures file line name of current level.
	 * 
	 * Caso o nmero da linha no esteja disponvel o valor 0  retornado.
	 * 
	 * @return integer
	 */
	public function getLine()
	{
		if ( array_key_exists( 'line', $this->aCurrent ) === false )
		{
			return 0;
		}
		return (integer) $this->aCurrent['line'];
	}
	
	/**
	 * Obtm o escopo do nvel corrente.
	 * 
	 * Captures scope of current level.
	 * 
	 * O tipo de escopo define o ambiente do nvel atual, ele pode ser funo,
	 * mtodo (funo de classe), esttico (mtodo esttico) ou arquivo. Esses
	 * tipos esto definidos nessa classe pelas constantes com prefixo "SCOPE_".
	 * Os possveis valores retornados so: {@link BackTrace::SCOPE_FILE},
	 * {@link BackTrace::SCOPE_FUNCTION}, {@link BackTrace::SCOPE_METHOD} e
	 * {@link BackTrace::SCOPE_STATIC}. Os escopos esto definidos de forma
	 * parecida com o ndice "type" da lista retornada pela funo
	 * {@link debug_backtrace()}, porm existe diferenciao entre o escopo
	 * funo e arquivo. Para obter o mesmo comportamento desse ndice "type"
	 * utilize {@link BackTrace::getType()}.
	 * 
	 * @return string
	 */
	public function getScope()
	{
		$sType = $this->getType();
		if ( $sType === '' )
		{
			$sType = $this->getFunction() === '' ? self::SCOPE_FILE : self::SCOPE_FUNCTION;
		}
		return $sType;
	}
	
	/**
	 * Obtm o tipo do nvel corrente.
	 * 
	 * Captures type of current level.
	 * 
	 * O tipo de escopo define o ambiente do nvel atual, ele pode ser mtodo
	 * (funo de classe), esttico (mtodo esttico) ou nenhum. Para obter um
	 * maior detalhamento do escopo utilize {@link BackTrace::getScope()}.
	 * 
	 * @return string
	 */
	public function getType()
	{
		if ( array_key_exists( 'type', $this->aCurrent ) === false )
		{
			return '';
		}
		return (string) $this->aCurrent['type'];
	}
	
	/**
	 * Move o ponteiro de rastreamento para o nvel mais baixo.
	 * 
	 * Moves trace pointer to last level.
	 * 
	 * @return void
	 */
	public function levelBottom()
	{
		$this->setLevel( count( $this->aStack ) - 1 );
	}
	
	/**
	 * Move o ponteiro de rastreamento para baixo.
	 * 
	 * Moves trace pointer to down.
	 * 
	 * Quando o valor retronado  verdadeiro a classe passa a manipular as
	 * informaes do nvel imediatamente inferior ao atual. Caso contrrio,
	 * quando o retorno  falso, o ponteiro no  alterado. A quantidade de
	 * nveis a ser abaixado  por padro o valor 1, porm ele pode ser
	 * redefinido pelo primeiro parmetro. Caso no seja possvel descer a
	 * quantidade de nveis desejado o segundo parmetro, que por padro 
	 * falso, indica se  para descer at onde for possvel.
	 * 
	 * @param integer $iDepth
	 * @param boolean $bTry caso o nvel no exista tentar at onde for possvel
	 * @return boolean
	 */
	public function levelDown( $iDepth = 1, $bTry = false )
	{
		$iDepth = (integer) $iDepth;
		$bReturn = $this->setLevel( $this->iLevel + $iDepth );
		if ( $bReturn === false && $bTry === true )
		{
			$iDepth > 0 ? $this->levelTop() : $this->levelBottom();
		}
		return $bReturn;
	}
	
	/**
	 * Move o ponteiro de rastreamento para o nvel mais alto.
	 * 
	 * Moves trace pointer to first level.
	 * 
	 * @return void
	 */
	public function levelTop()
	{
		$this->setLevel( 0 );
	}
	
	/**
	 * Move o ponteiro de rastreamento para cima.
	 * 
	 * Moves trace pointer to up.
	 * 
	 * Quando o valor retronado  verdadeiro a classe passa a manipular as
	 * informaes do nvel imediatamente superior ao atual. Caso contrrio,
	 * quando o retorno  falso, o ponteiro no  alterado. A quantidade de
	 * nveis a ser subido  por padro o valor 1, porm ele pode ser redefinido
	 * pelo primeiro parmetro. Caso no seja possvel subir a quantidade de
	 * nveis desejado o segundo parmetro, que por padro  falso, indica se 
	 * para subir at onde for possvel.
	 * 
	 * @param integer $iDepth
	 * @param boolean $bTry caso o nvel no exista tentar at onde for possvel
	 * @return boolean
	 */
	public function levelUp( $iDepth = 1, $bTry = false )
	{
		$iDepth = (integer) $iDepth;
		$bReturn = $this->setLevel( $this->iLevel - $iDepth );
		if ( $bReturn === false && $bTry === true )
		{
			$iDepth > 0 ? $this->levelBottom() : $this->levelTop();
		}
		return $bReturn;
	}
	
	/**
	 * Move o ponteiro de rastreamento.
	 * 
	 * Moves trace pointer.
	 * 
	 * Quando o valor retronado  verdadeiro a classe passa a manipular as
	 * informaes do nvel indicado pelo parmetro. Caso contrrio, quando o
	 * retorno  falso, o ponteiro no  alterado.
	 * 
	 * @param integer $iLevel nvel do rastreamento
	 * @return boolean
	 */
	public function setLevel( $iLevel )
	{
		$iLevel = (integer) $iLevel;
		if ( array_key_exists( $iLevel, $this->aStack ) === true )
		{
			$this->iLevel = $iLevel;
		}
		$this->aCurrent = &$this->aStack[$this->iLevel];
		return $this->iLevel === $iLevel;
	}
	
}
